from typing import Tuple, Union
from pathlib import Path
import numpy as np
import multiresolutionimageinterface as mir
from .interface import Reader, num


class AsapSlide(Reader):
    def __init__(self, path: Union[str, Path]):
        self.slide = mir.MultiResolutionImageReader().open(path)

    @property
    def level_count(self) -> int:
        return self.slide.getNumberOfLevels()

    @property
    def base_mpp(self) -> float:
        return float(self.slide.getProperty('openslide.mpp-x'))

    def dimension(self, level: int = 0) -> Tuple[int, int]:
        return self.slide.getLevelDimensions(level % self.level_count)

    def downsample(self, level: int = 0) -> float:
        return self.slide.getLevelDownsample(level % self.level_count)

    def region(self, level: int, left: num, up: num, right: num, down: num) -> np.ndarray:
        downsample = self.downsample(level)
        l0 = round(left * downsample)
        u0 = round(up * downsample)
        w = round(right - left)
        h = round(down - up)
        patch = self.slide.getUCharPatch(startX=l0, startY=u0, width=w, height=h, level=level)
        return patch



import io
from typing import Tuple, Union

from .definition import Shape


class Full(Shape):

    __slots__ = ()

    def is_valid(self):
        return True

    def buffer(self, distance: float):
        return self

    def standard(self):
        return self

    @property
    def geo(self):
        return None

    def clean(self):
        return self

    def offset(self, pos: Union[complex, Tuple[float, float]]):
        return self

    def scale(self, ratio: float, origin=0j):
        return self

    def rotate(self, degree: float, origin=0j):
        return self

    def flip_x(self, a: float):
        return self

    def flip_y(self, b: float):
        return self

    def flip(self, a: float, b: float):
        return self

    def is_joint(self, other) -> bool:
        return other is not Shape.EMPTY

    def if_contain(self, other) -> bool:
        return other is not Shape.EMPTY

    def inter(self, other):
        return other

    def union(self, other):
        return self

    def diff(self, other):
        return -other

    def merge(self, other):
        return self

    def remove(self, other):
        return -other

    def simplify(self, tolerance: float):
        return self

    def smooth(self, distance: float):
        return self

    @property
    def convex(self):
        return self

    @property
    def mini_rect(self):
        return self

    @property
    def region(self):
        return self

    @property
    def center(self) -> Tuple[int, int]:
        return 0, 0

    @property
    def area(self) -> float:
        return 0

    @property
    def perimeter(self) -> float:
        return 0

    @property
    def bounds(self) -> Tuple[int, int, int, int]:
        return 0, 0, 0, 0

    @property
    def outer(self):
        return self

    @property
    def inner(self):
        return self

    def sep_in(self):
        return [Shape.FULL], []

    def sep_out(self):
        return [Shape.FULL]

    def sep_p(self):
        raise[]

    def copy(self):
        return self

    def comp(self):
        return Shape.EMPTY

    @property
    def reversed(self) -> bool:
        return True

    def dumps(self) -> str:
        return 'Full'

    def dumpb(self, f: io.BufferedWriter):
        return 'Full', None, None


Shape.FULL = Full()



import os.path

import nbformat


def ipynb2pycode(ipynb: str) -> str:
    # 可以输入 path，也可以直接输入文件内容
    if os.path.exists(ipynb):
        nb = nbformat.read(ipynb, as_version=4)
    else:
        nb = nbformat.reads(ipynb, as_version=4)
    lines = []
    for cell in nb.cells:
        if cell.cell_type == 'code':
            code = cell.source
            for line in code.splitlines():
                stripped = line.strip()
                if not stripped.startswith('#') and stripped != '':
                    lines.append(line)
    return '\n'.join(lines)



import sys
import os.path
from typing import Union
from pathlib import Path
import urllib.request
import cv2
import numpy as np
import onnxruntime


# 来自 modnet 的权重
# 仓库地址：https://github.com/ZHKKKe/MODNet
# 权重地址：https://drive.google.com/drive/folders/1umYmlCulvIFNaqPjwod1SayFmSRHziyR
# 使用 Apache 2.0 协议
# 谷歌云使用动态链接，所以此处将权重上传至 github 仓库
target_uri = 'https://raw.githubusercontent.com/name-used/jassor/refs/heads/master/resources/modnet_photographic_portrait_matting.onnx'


def process(image: np.ndarray, onnx_path: Union[str, Path] = './modnet_human_detect.onnx') -> np.ndarray:
    if not os.path.exists(onnx_path):
        download_checkpoint(onnx_path)
    session = onnxruntime.InferenceSession(onnx_path, None)
    input_name = session.get_inputs()[0].name
    output_name = session.get_outputs()[0].name
    result = session.run([output_name], {input_name: transpose(image)})
    matte = (np.squeeze(result[0]) * 255).astype('uint8')
    im_h, im_w = image.shape[:2]
    matte = cv2.resize(matte, dsize=(im_w, im_h), interpolation=cv2.INTER_AREA)
    return matte


def transpose(image: np.ndarray):
    # unify image channels to 3
    if len(image.shape) == 2:
        image = image[:, :, None]
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
    elif image.shape[2] == 4:
        image = image[:, :, 0:3]

    # normalize values to scale it between -1 to 1
    image = (image - 127.5) / 127.5

    im_h, im_w, im_c = image.shape
    x, y = get_scale_factor(im_h, im_w, ref_size=512)

    # resize image
    image = cv2.resize(image, None, fx=x, fy=y, interpolation=cv2.INTER_AREA)

    # prepare input shape
    image = np.transpose(image)
    image = np.swapaxes(image, 1, 2)
    image = np.expand_dims(image, axis=0).astype('float32')
    return image


# Get x_scale_factor & y_scale_factor to resize image
def get_scale_factor(im_h, im_w, ref_size):

    if max(im_h, im_w) < ref_size or min(im_h, im_w) > ref_size:
        if im_w >= im_h:
            im_rh = ref_size
            im_rw = int(im_w / im_h * ref_size)
        else:
            im_rw = ref_size
            im_rh = int(im_h / im_w * ref_size)
    else:
        im_rh = im_h
        im_rw = im_w

    im_rw = im_rw - im_rw % 32
    im_rh = im_rh - im_rh % 32

    x_scale_factor = im_rw / im_w
    y_scale_factor = im_rh / im_h

    return x_scale_factor, y_scale_factor


def download_checkpoint(path: Union[str, Path]):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    try:
        sys.stderr.write(f'start download modnet weights from {target_uri}\n')
        sys.stderr.write(f'for more information see https://github.com/ZHKKKe/MODNet\n')
        urllib.request.urlretrieve(target_uri, path)
        sys.stderr.write(f'download success in path {path}\n')
    except Exception as e:
        sys.stderr.write(f'modnet weights download failed, please check network\n')



import math
from typing import List


def uniform_iter(T_len: int, I_len: int, max_gap: int) -> List[int]:
    """
    用于覆盖指定长度区间的函数，生成均匀分布的最小覆盖区间。

    参数：
    T_len (int): 需要被覆盖的总区间长度。例如 100。
    I_len (int): 每个覆盖区间的长度。例如 5。
    max_gap (int): 相邻覆盖区间起始位置的最大间距，确保覆盖的连续性。例如 3。

    返回：
    list: 包含每个覆盖区间起始位置的整数列表，以满足指定条件。
    """
    if I_len > T_len:
        return []
    
    remainder = (I_len - T_len) % max_gap
    if remainder == 0:
        return list(range(0, T_len-I_len + 1, max_gap))

    count = math.ceil((T_len - I_len) / max_gap)
    step = (T_len - I_len) / count

    # 步长数 + 原窗口0计数 => 总迭代量
    return [round(step * i) for i in range(count + 1)]



from .interface import Reader
from .reader import load
from .single_predict_crop_dataset import SingleDataset
from .utils import trans_norm, trans_linear, sample_image, sample_slide


__all__ = [
    'Reader',
    'load',
    'SingleDataset',
    'trans_norm',
    'trans_linear',
    'sample_image',
    'sample_slide',
]



import importlib
my_modules = {
    'Marker': '.blind_mark',
    'Masking': '.create_mask',
    'data': '.data',
}


def __getattr__(name):
    if name in my_modules:
        module = importlib.import_module(my_modules[name], __package__)
        return module
    else:
        raise ModuleNotFoundError(f'The import name {name} not in this components, check if want to import {list(my_modules.keys())}')


__all__ = list(my_modules)



from .definition import Shape, Single as SingleShape, Multi as MultiShape, ConvertMulti2SingleException
from .impl_multi_simple import MultiSimplePolygon
from .impl_multi_complex import MultiComplexPolygon
from .impl_region import Region
from .impl_circle import Circle
from .impl_single_simple import SimplePolygon
from .impl_single_complex import ComplexPolygon
from .impl_empty import Empty
from .impl_full import Full
from .polygon_creators import create_polygon, create_triangle, create_regular_polygon, create_sector
from .functional import load, loads, loadb


EMPTY = Shape.EMPTY
FULL = Shape.FULL

__all__ = [
    'Shape',
    'SingleShape',
    'Region',
    'Circle',
    'SimplePolygon',
    'ComplexPolygon',
    'MultiShape',
    'MultiSimplePolygon',
    'MultiComplexPolygon',
    'EMPTY',
    'FULL',
    'create_polygon',
    'create_triangle',
    'create_regular_polygon',
    'create_sector',
    'ConvertMulti2SingleException',
    'load',
    'loads',
    'loadb',
]

'''
关于 shapely 中的类型:
'Point',
'LineString',
'LinearRing',
'Polygon',
'MultiPoint',
'MultiLineString',
'MultiPolygon',
'GeometryCollection',
'''



import numpy as np
from scipy.fftpack import dct, idct

k = 8


def map_to(img_in: np.ndarray):
    """
    将给定图像映射到目标域上
    :param img_in: uint8 图像
    :return: dct 图像
    """
    h, w = img_in.shape
    H = h + -h % k
    W = w + -w % k
    image = np.zeros(shape=(H, W), dtype=np.uint8)
    image[:h, :w] = img_in
    # 数据空间变换 [0, 255] -> [0.5, 255.5] -> (-1, 1) -> (-∞, +∞)
    image = (image.astype(np.float64) + 0.5 - 128) / 128
    image = np.tan(image * np.pi / 2)

    # 对每个块进行 DCT 变换, 变换中留存主值
    result = np.zeros_like(image, dtype=np.float32)
    for i in range(0, H, k):
        for j in range(0, W, k):
            block = image[i:i + k, j:j + k]
            block = dct(dct(block.T, norm='ortho').T, norm='ortho')
            # block = u @ block @ v
            result[i:i + k, j:j + k] = block
    return result


def imap_to(img_in: np.ndarray):
    """
    将给定目标域图像映射回 uint8
    :param img_in: dct 图像
    :return: uint8 图像
    """
    h, w = img_in.shape
    H = h + -h % k
    W = w + -w % k
    image = np.zeros(shape=(H, W), dtype=np.float32)
    image[:h, :w] = img_in
    result = np.zeros_like(image, dtype=np.float32)
    for i in range(0, H, k):
        for j in range(0, W, k):
            block = image[i:i + k, j:j + k]
            # block = u.T @ block @ v.T
            block = idct(idct(block.T, norm='ortho').T, norm='ortho')
            result[i:i + k, j:j + k] = block

    # 数据空间变换 [0, 255] -> [0.5, 255.5] -> (-1, 1) -> (-∞, +∞)
    result = np.arctan(result) * 2 / np.pi
    result = (result * 128 + 128).astype(np.uint8)
    return result



from scipy import signal
from scipy import ndimage
import cv2
import numpy as np


def process(image: np.ndarray, blur_size: int = 9, thresh_min: float = 0, thresh_max: float = 1.):
    # unify image channels to 3
    if len(image.shape) == 2:
        image = image[:, :, None]
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
    elif image.shape[2] == 4:
        image = image[:, :, 0: 3]
    image = image / 255
    blurred = cv2.GaussianBlur(image, (blur_size, blur_size), sigmaX=0, sigmaY=0)
    divided = (image + 0.01) / (blurred + 0.01)
    mask = divided
    # mask = np.where(divided < 1, divided, 1 / divided)
    mask = mask.clip(thresh_min, thresh_max)
    mask -= thresh_min
    mask /= thresh_max - thresh_min
    mask = (mask * 255).round().astype(np.uint8)

    return mask



from typing import List, Tuple

import shapely
from shapely.geometry.base import BaseGeometry

from .definition import Shape, Single, CoordinatesNotLegalException, NoParametersException
from .impl_single_complex import ComplexPolygon
from . import functional as F


class SimplePolygon(ComplexPolygon):
    """
    单-单连通多边形, 创建方式有三:
    1. 指定 outer
    2. 指定 geo
    3. 指定 single
    遵循逆序优先规则
    当 polygon 是 ComplexPolygon 时, 忽视其内轮廓
    """

    __slots__ = ()

    def __init__(
            self,
            outer: List[Tuple[float, float]] = None,
            geo: BaseGeometry = None,
            single: Single = None,
            from_p: List[Tuple[float, float]] = None,
            reverse: bool = False
    ):
        if geo is not None:
            assert isinstance(geo, shapely.Polygon), 'geo 必须是 Polygon'
            assert geo.boundary.geom_type.upper() == 'LINESTRING', 'geo 必须是单连通的'
        elif single is not None:
            assert isinstance(single, Single), 'Multi 类型无法转换为 Single'
            geo = single.geo
        elif from_p is not None:
            outer = from_p
            geo = shapely.Polygon(shell=outer, holes=[])
        elif outer is not None:
            # 对用户输入进行检查和修复
            geo = shapely.Polygon(shell=outer, holes=[])
            geo = F.norm_geo(geo)
            # 创建时要求轮廓必须合法
            if geo is None or isinstance(geo, shapely.MultiPolygon):
                raise CoordinatesNotLegalException(f'creating single polygon with geo=={type(geo)}')
            # 还得是单连通的
            if bool(geo.interiors):
                raise CoordinatesNotLegalException(f'SimplePolygon not allowed interiors coordinates with geo interiors nums: {len(geo.interiors)}')
        else:
            # 没有任何参数的话，要报个错
            raise NoParametersException(f'Any of such parameters have to be provided: outer, geo, single, from_p')
        super().__init__(None, geo=geo, single=None, from_p=None, reverse=reverse)

    @property
    def outer(self) -> Single:
        # 外轮廓(正形)
        return Single.SIMPLE(geo=self.geo)

    @property
    def inner(self) -> Shape:
        # 内轮廓(负形)
        return Shape.EMPTY

    def sep_out(self) -> List[Single]:
        # 外分解
        return [Single.SIMPLE(geo=self.geo)]

    def sep_p(self) -> List[Tuple[float, float]]:
        # 点分解
        outer = list(shapely.get_exterior_ring(self.geo).coords)
        return outer

    @property
    def cls(self) -> type:
        return SimplePolygon


Single.SIMPLE = SimplePolygon



import cv2
import numpy as np
from skimage import measure
from skimage.measure import regionprops
from copy import deepcopy
from scipy.signal import argrelextrema
from skimage.filters.rank import entropy
from skimage.morphology import disk


def process(image: np.ndarray):
    if len(image.shape) == 2:
        pass
    if image.shape[2] == 1:
        image = image[:, :, 0]
    elif image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    elif image.shape[2] == 4:
        image = cv2.cvtColor(image, cv2.COLOR_RGBA2GRAY)

    source = deepcopy(image)
    ent = entropy(source, disk(5))
    hist = list(np.histogram(ent, 30))
    minindex = list(argrelextrema(hist[0], np.less))

    thresh_localminimal = 0
    for i in range(len(minindex[0])):
        temp_thresh = hist[1][minindex[0][i]]
        if 1 < temp_thresh < 4:
            thresh_localminimal = temp_thresh

    eimage = entropy(image, disk(3))
    new_picture = np.ndarray(shape=eimage.shape)  # [[False] * image.shape[1]] * image.shape[0]
    for rn, row in enumerate(eimage):
        for pn, pixel in enumerate(row):
            if pixel < thresh_localminimal:
                new_picture[rn, pn] = True
            else:
                new_picture[rn, pn] = False
    entropy_image = new_picture.astype('b')

    thresh1 = (255 * entropy_image).astype('uint8')
    mask_255 = cv2.bitwise_not(deepcopy(thresh1))

    redolbl = measure.label(np.array(mask_255), connectivity=2)
    redprops = regionprops(redolbl)

    redAreaData = []
    for r in range(len(redprops)):
        redAreaData.append(redprops[r].area)

    # 新方法
    redAreaData = sorted(enumerate(redAreaData), key=lambda p: p[1], reverse=True)
    # choses = [i for i, v in redAreaData]
    choses = []
    for i, v in redAreaData:
        if v > 0 and v / sum(p[1] for p in redAreaData[:i+1]) > 0.005:
            choses.append(i)
        else:
            break
    max_mask = np.zeros_like(image)
    for chose in choses:
        max_mask[redolbl == chose + 1] = 255

    return max_mask



import math
from typing import Union

import PIL.Image
import numpy as np
import torch
import jassor.utils as J
from .interface import Reader

std = np.asarray([0.229, 0.224, 0.225])
mean = np.asarray([0.485, 0.456, 0.406])


def trans_norm(target_input: Union[np.ndarray, torch.Tensor], channel_dim: int = 1):
    shape = [3 if i == channel_dim else 1 for i in range(len(target_input.shape))]
    m = mean.reshape(shape)
    s = std.reshape(shape)
    return (target_input / 255 - m) / s


def trans_linear(target_input: Union[np.ndarray, torch.Tensor]):
    return (target_input / 255 - 0.5) / 0.5


def sample_image(image: Union[np.ndarray, PIL.Image.Image], kernel_size: int, step: int):
    w, h = image.shape[:2][::-1] if isinstance(image, np.ndarray) else image.size
    k, s = kernel_size, step
    # image 的 level 恒为 0
    return [(0, x, y, x+k, y+k) for y in J.uniform_iter(h, k, s) for x in J.uniform_iter(w, k, s)]


def sample_slide(reader: Reader, level: int, kernel_size: int, step: int, mask: np.ndarray = None):
    W, H = reader.dimension(level)
    k, s = kernel_size, step
    if mask is not None:
        basic_samples = [(x, y, x+k, y+k) for y in J.uniform_iter(H, k, s) for x in J.uniform_iter(W, k, s)]
        h, w = mask.shape[:2]
        mask_samples = [[math.floor(l*w/W), math.floor(u*h/H), math.ceil(r*w/W), math.ceil(d*h/H)] for l, u, r, d in basic_samples]
        filtered_samples = [
            (level, l, u, r, d) for (l, u, r, d), (ml, mu, mr, md) in zip(basic_samples, mask_samples)
            if mask[mu: md, ml: mr].any()
        ]
        return filtered_samples
    else:
        return [(level, x, y, x+k, y+k) for y in J.uniform_iter(H, k, s) for x in J.uniform_iter(W, k, s)]



import abc
from typing import Tuple, Union
import numpy as np

num = Union[float, int]


class Reader:
    @property
    @abc.abstractmethod
    def level_count(self) -> int:
        raise NotImplemented

    @property
    @abc.abstractmethod
    def base_mpp(self) -> float:
        raise NotImplemented

    def mpp(self, level: int = 0) -> float:
        return self.base_mpp * self.downsample(level)

    @abc.abstractmethod
    def dimension(self, level: int = 0) -> Tuple[int, int]:
        raise NotImplemented

    @abc.abstractmethod
    def downsample(self, level: int = 0) -> float:
        raise NotImplemented

    @abc.abstractmethod
    def region(self, level: int, left: num, up: num, right: num, down: num) -> np.ndarray:
        raise NotImplemented

    def thumb(self, level: int = -1) -> np.ndarray:
        level = level % self.level_count
        w, h = self.dimension(level)
        return self.region(level, 0, 0, w, h)


# class Dataset(torch.utils.data.Dataset, abc.ABC):
#     def __init__(self, source: Any):
#         super().__init__()
#         self.source = source
#         self.samples = []
#
#     def __len__(self):
#         return len(self.samples)
#
#     def __getitem__(self, item):
#         return self.load(self.samples[item])
#
#     def __iter__(self):
#         for i in range(len(self)):
#             yield self[i]
#
#     @abc.abstractmethod
#     def load(self, sample) -> Any: pass



from typing import Union
import numpy as np
import torch


class BBox:
    XYWH = 'cxcywh'
    LURD = 'lurd'
    LUWH = 'luwh'

    def __init__(self, bbox: Union[list, np.ndarray, torch.Tensor], box_format: str = LURD):
        # bbox = array: [(a, b, c, d)]
        # abcd 的含义与 format 有关
        self._bbox = np.asarray(bbox, dtype=np.float32)
        self._box_format = box_format

    @property
    def bbox(self):
        return self._bbox

    @property
    def contour(self):
        return bbox_to_contour(self._bbox)

    def join_region(self, l, u, r, d, eps=5):
        bbox = self.lurd()
        bbox, ps = bbox_join_region(bbox, l, u, r, d, eps)
        return BBox(bbox, self.LURD), ps

    def area(self):
        bbox = self.xywh()
        return bbox._bbox[2] * bbox._bbox[3]

    def inter(self, bbox):
        assert isinstance(bbox, BBox), 'Type BBox is Necessary'
        bbox1 = self.xywh()._bbox
        bbox2 = bbox.xywh()._bbox
        inter_box, select_1, select_2 = bbox_inter(bbox1, bbox2)
        return BBox(inter_box, self.LURD), select_1, select_2

    def xywh(self):
        if self._box_format == self.XYWH:
            return self
        elif self._box_format == self.LUWH:
            bbox = bbox_luwh2xywh(self._bbox)
            return BBox(bbox, self.XYWH)
        elif self._box_format == self.LURD:
            bbox = bbox_lurd2xywh(self._bbox)
            return BBox(bbox, self.XYWH)
        else:
            raise TypeError('bbox format not implemented')

    def lurd(self):
        if self._box_format == self.LURD:
            return self
        elif self._box_format == self.XYWH:
            bbox = bbox_xywh2lurd(self._bbox)
            return BBox(bbox, self.LURD)
        elif self._box_format == self.LUWH:
            bbox = bbox_luwh2lurd(self._bbox)
            return BBox(bbox, self.LURD)
        else:
            raise TypeError('bbox format not implemented')

    def luwh(self):
        if self._box_format == self.LUWH:
            return self
        elif self._box_format == self.XYWH:
            bbox = bbox_xywh2luwh(self._bbox)
            return BBox(bbox, self.LUWH)
        elif self._box_format == self.LURD:
            bbox = bbox_lurd2luwh(self._bbox)
            return BBox(bbox, self.LUWH)
        else:
            raise TypeError('bbox format not implemented')


def bbox_to_contour(bbox: np.ndarray):
    # 使用 lurd 格式
    l, u, r, d = bbox.T
    contour = [[l, u], [r, u], [r, d], [l, d], [l, u]]
    contour = np.asarray(contour).transpose((2, 0, 1))
    return contour


def bbox_join_region(bbox: np.ndarray, l, u, r, d, eps):
    # join_region 使用 lurd 格式
    bbox = bbox.copy()
    bbox[bbox[:, 0] >= r - eps, 0] = np.nan  # bbox at box right
    bbox[bbox[:, 1] >= d - eps, 1] = np.nan  # bbox at box down
    bbox[bbox[:, 2] <= l + eps, 2] = np.nan  # bbox at box left
    bbox[bbox[:, 3] <= u + eps, 3] = np.nan  # bbox at box up
    ps = ~np.isnan(bbox.sum(axis=0))
    bbox = bbox[ps, :]
    ps = ps.nonzero()

    # limit to box
    bbox[bbox[:, 0] < l, 0] = l
    bbox[bbox[:, 1] < u, 1] = u
    bbox[bbox[:, 2] > r, 2] = r
    bbox[bbox[:, 3] > d, 3] = d

    # [l, u, r, d]
    return bbox, ps


def bbox_inter(bbox1: np.ndarray, bbox2: np.ndarray):
    # inter 先使用 xywh 格式，再使用 lurd 格式，返回的是 lurd 格式
    # 计算相交面积，面积非零的索引拎出来
    area_matrix = bbox_inter_area_matrix(bbox1, bbox2)
    i, j = area_matrix.nonzero()
    # 按索引拎出来之后，轮廓之间就对齐了
    bbox1 = bbox1[i, :]
    bbox2 = bbox2[j, :]
    # 转换格式
    bbox1 = bbox_xywh2lurd(bbox1)
    bbox2 = bbox_xywh2lurd(bbox2)
    # 然后按照对齐的结果计算轮廓匹配结果
    l1, u1, r1, d1 = bbox1.T
    l2, u2, r2, d2 = bbox2.T
    l, u, r, d = np.maximum(l1, l2), np.maximum(u1, u2), np.minimum(r1, r2), np.minimum(d1, d2)
    return np.stack([l, u, r, d], axis=1), i, j


def bbox_inter_area_matrix(bbox1: np.ndarray, bbox2: np.ndarray) -> np.ndarray:
    # inter 使用 xywh 格式
    # 逻辑：对于 bbox1[i] 和 bbox2[j]，做如下计算：
    # if abs(xi - xj) * 2 >= (wi + wj): continue
    # if abs(yi - yj) * 2 >= (hi + hj): continue
    # 我要计算的不是两个轮廓的面积交并比，而是上级轮廓占下级的比例
    # x 方向上相交区域为 min(max(0, wi+wj-abs(xi-xj)), min(2*wi, 2*wj))，y 方向同理
    # inter = min((wi + wj) / 2 - abs(xi - xj), wi, wj) * min((hi + hj) / 2 - abs(yi - yj), hi, hj)
    # 然后把这个逻辑直接放到 numpy 上加速计算
    # 按列拆分
    x1, y1, w1, h1 = bbox1.T
    x2, y2, w2, h2 = bbox2.T
    # 计算中间矩阵（1 在 行，2 在 列）
    x_sub = abs(x1[:, None] - x2[None, :])
    y_sub = abs(y1[:, None] - y2[None, :])
    w_add = w1[:, None] + w2[None, :]
    h_add = h1[:, None] + h2[None, :]
    disjoint = (x_sub * 2 > w_add) | (y_sub * 2 > h_add)
    inter_w = w_add / 2 - x_sub
    inter_w = np.minimum(inter_w, w1[:, None])
    inter_w = np.minimum(inter_w, w2[None, :])
    inter_h = h_add / 2 - x_sub
    inter_h = np.minimum(inter_h, h1[:, None])
    inter_h = np.minimum(inter_h, h2[None, :])
    inter_area = (~disjoint) * inter_w * inter_h
    return inter_area


def bbox_lurd2xywh(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (2, 3)] = bbox[:, (2, 3)] - bbox[:, (0, 1)]
    bbox[:, (0, 1)] = bbox[:, (0, 1)] + bbox[:, (2, 3)] / 2
    return bbox


def bbox_xywh2lurd(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (0, 1)] = bbox[:, (0, 1)] - bbox[:, (2, 3)] / 2
    bbox[:, (2, 3)] = bbox[:, (0, 1)] + bbox[:, (2, 3)]
    return bbox


def bbox_luwh2xywh(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (0, 1)] = bbox[:, (0, 1)] + bbox[:, (2, 3)] / 2
    return bbox


def bbox_xywh2luwh(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (0, 1)] = bbox[:, (0, 1)] - bbox[:, (2, 3)] / 2
    return bbox


def bbox_lurd2luwh(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (2, 3)] = bbox[:, (2, 3)] - bbox[:, (0, 1)]
    return bbox


def bbox_luwh2lurd(bbox: np.ndarray) -> np.ndarray:
    bbox = bbox.copy()
    bbox[:, (2, 3)] = bbox[:, (0, 1)] + bbox[:, (2, 3)]
    return bbox



import numpy as np


def cut_label(label: np.ndarray, l: int, u: int, r: int, d: int, ignore: int = 5, MIN: int = -999_999_999) -> np.ndarray:
    # label = array([(l, u, r, d)], dtype=(float, int))
    # assert label > 0
    if len(label) == 0: return label

    label = label.copy()

    # choose in box
    label[label[:, 0] >= r - ignore, 0] = MIN  # label at box right
    label[label[:, 1] >= d - ignore, 1] = MIN  # label at box down
    label[label[:, 2] <= l + ignore, 2] = MIN  # label at box left
    label[label[:, 3] <= u + ignore, 3] = MIN  # label at box up
    label = label[label.sum(axis=1) > 0, :]

    # limit to box
    label[label[:, 0] < l, 0] = l
    label[label[:, 1] < u, 1] = u
    label[label[:, 2] > r, 2] = r
    label[label[:, 3] > d, 3] = d

    # absolute -> relative
    label[:, (0, 2)] -= l
    label[:, (1, 3)] -= u

    # [t, l, u, r, d] :: [0 - w/h]
    return label



import math
from typing import Iterable
from .definition import Shape
from .impl_single_simple import SimplePolygon
from .impl_circle import Circle


def create_regular_polygon(n: int, len_side: float = None, center_radius: float = None) -> Shape:
    """
    创建正 n 边形，优先使用边长，未提供边长时使用外接圆半径，二者均不提供时，默认生成边长为 1 的正 n 边形
    # 生成的 n 边形总是中心在原点，且一个顶点在 x 轴上
    生成的三角形第一个顶点总是原点，第一条边总是在 x 轴上
    :param n: 边数
    :param len_side: 边长
    :param center_radius: 外接圆半径
    :return: 一个定义为正 n 边形的 Shape
    """
    if len_side:
        # 换算式由 GPT 提供：
        # calc_side_length = 2 * radius * np.sin(np.pi / n)
        # radius = side_length / (2 * np.sin(np.pi / n))
        r = len_side / (2 * math.sin(math.pi / n))
    else:
        r = center_radius or 1
    base_directs = [2 * math.pi / n * i for i in range(n)]
    base_points = [(math.cos(d), math.sin(d)) for d in base_directs]
    outer = [(p * r, q * r) for p, q in base_points]
    # return SimplePolygon(outer=outer)
    shape = SimplePolygon(outer=outer)
    shape -= (r, 0)
    shape.rotate(-90 - 360 / n / 2, origin=(0, 0))
    return shape


def create_triangle(len_sides: Iterable[float], degrees: Iterable[float] = None) -> Shape:
    """
    创建三角形，边长与角度不必全给，只需给够满足确定一个三角形的最小需求即可
    边角序列对应的几何关系是：边1、角1、边2、角2、边3、角3、边1、...
    生成的三角形第一个顶点总是原点，第一条边总是在 x 轴上
    :param len_sides: 边长序列，至少要给一个长度，缺失的边长可以用 None 占位
    :param degrees: 角度序列，若三边长度全给，则可以不提供角度 —— 采用角度制，0 < degree < 180，缺失的角度可以用 None 占位
    :return:
    """
    # 参数标准化
    sides = list(len_sides)
    sides.extend([None] * max(0, 3 - len(sides)))
    dgs = list(degrees or [])
    dgs.extend([None] * max(0, 3 - len(dgs)))
    # 补全三边三角
    sides, dgs = _triangle_complete_arguments(sides, dgs)

    # 几何学角度转化为坐标系倾角
    dgs = [180 - d for d in dgs]
    dgs = [sum(dgs[:i + 1]) for i, _ in enumerate(dgs)]
    assert abs(dgs[-1] - 360) < 0.05, f'角度差过大，请检查输入参数: {len_sides}, {degrees}'

    # 创建三角形
    x1 = y1 = 0
    x2 = sides[0]
    y2 = 0
    # 只有第三个顶点需要计算
    x3 = x2 + math.cos(dgs[0] * math.pi / 180) * sides[1]
    y3 = y2 + math.sin(dgs[0] * math.pi / 180) * sides[1]
    # 但还需要计算第四个顶点，以检查所得形状是否符合要求
    x4 = x3 + math.cos(dgs[1] * math.pi / 180) * sides[2]
    y4 = y3 + math.sin(dgs[1] * math.pi / 180) * sides[2]
    assert abs(x4) + abs(y4) < max(s for s in len_sides if s is not None) * 0.05, f'计算结果不构成三角形，请检查输入参数: {len_sides}, {degrees}'

    return SimplePolygon([(x1, y1), (x2, y2), (x3, y3)])


def _triangle_complete_arguments(sides, dgs):
    if dgs.count(None) <= 1:
        # 有两角，第三角可以直接计算出来
        if dgs.count(None) == 1:
            dgs[dgs.index(None)] = 180 - sum(d for d in dgs if d)
        # 然后用正弦定理补边长
        p = [s is not None for s in sides].index(True)
        for i, s in enumerate(sides):
            if s is not None: continue
            sides[i] = sides[p] / math.sin(dgs[(p + 1) % 3] * math.pi / 180) * math.sin(dgs[(i + 1) % 3] * math.pi / 180)
        return sides, dgs
    # 否则至少有两边
    if sides.count(None) == 1:
        # SSA 不能唯一确定三角形，因此只考虑 SAS，条件不满足直接抛错不管
        assert dgs[(sides.index(None) + 1) % 3] is not None, f'请使用 SSS、AAS、SAS 的方式定义三角形，检查输入参数：{sides}, {dgs}'
        # 正对角有值，用余弦定理补边长
        p = [s is not None for s in sides].index(False)
        sides[p] = (
            sides[(p + 1) % 3] ** 2 +
            sides[(p - 1) % 3] ** 2 -
            2 * sides[(p + 1) % 3] * sides[(p - 1) % 3] * math.cos(
               dgs[(p + 1) % 3] * math.pi / 180
            )
        ) ** 0.5
    # 正弦定理求角度不可靠（钝角锐角问题），因此统一用余弦定理求角度
    for i, d in enumerate(dgs):
        if d is not None: continue
        dgs[i] = math.acos(
            (
               sides[i] ** 2 +
               sides[(i + 1) % 3] ** 2 -
               sides[(i - 1) % 3] ** 2
            ) / (
                2 * sides[i] * sides[(i + 1) % 3]
            )
        ) * 180 / math.pi
    return sides, dgs


def create_polygon(len_sides: Iterable[float], degrees: Iterable[float], ring_close: bool = True) -> Shape:
    """
    由于任意多边形不能简单由边长信息确定，所以必须给全边长和角度的参数
    :param len_sides:   边长数组
    :param degrees:     角度数组
    :param ring_close:  描述所给轮廓是否包含最后一条边和最后一个角
                    默认为真，为真时，依路径得到的最后一个点应当和起始点重合
                    为假时，依路径得到的最后一个点应当与起点不重合
    :return:
    """
    # 几何学角度转化为坐标系倾角
    dgs = [180 - d for d in degrees]
    dgs = [sum(dgs[:i+1]) for i, _ in enumerate(dgs)]
    if ring_close:
        assert abs(dgs[-1] - 360) < 5, f'角度差过大，请检查输入参数: {len_sides}, {degrees}'
    dgs.insert(0, 0)

    points = [(0, 0)]
    for s, d in zip(len_sides, dgs):
        x0, y0 = points[-1]
        x = x0 + math.cos(d * math.pi / 180) * s
        y = y0 + math.sin(d * math.pi / 180) * s
        points.append((x, y))

    if ring_close:
        assert abs(max(points[-1])) < max(len_sides) * 0.05, f'计算结果不符合预期，请检查输入参数: {len_sides}, {degrees}'
        return SimplePolygon(points[:-1])
    else:
        return SimplePolygon(points)


def create_sector(radius: float, degree: float, num: int = 100) -> Shape:
    """
    创建一个扇形
    :param radius:  扇形所对应圆的半径
    :param degree:  扇形所对应圆心角
    :param num:     描述扇形所用轮廓点数
    :return:
    """
    # 入参矫正检查
    if abs(degree) < 0.05: return Shape.EMPTY
    if abs(degree) >= 360: return Circle(0, 0, radius, num)
    # 角度值换弧度制
    angle = degree * math.pi / 180
    base_directs = [angle / num * i for i in range(num + 1)]
    base_points = [(math.cos(d), math.sin(d)) for d in base_directs]
    outer = [(0., 0.)] + [(p * radius, q * radius) for p, q in base_points]
    return SimplePolygon(outer=outer)



import importlib
my_modules = {
    'Logger': '.logger',
    'TimerManager': '.timer',
    'Timer': '.timer',
    'Queue': '.multiprocess',
    'Closed': '.multiprocess',
    'Process': '.multiprocess',
    'QueueMessageException': '.multiprocess',
    'JassorJsonEncoder': '.json_encoder',
    'Merger': '.merger',
    'random_colors': '.color',
    'random_rainbow_curves': '.color',
    'plot': '.jassor_plot_lib',
    'plots': '.jassor_plot_lib',
    'Table': '.table',
    'uniform_iter': '.iter_method',
    'crop': '.cropper',
    # 'SlideWriter': '.writer_asap',
    'SlideWriter': '.writer_tifffile',
    'image2slide': '.writer_tiff',
    'BBox': '.bbox',
    'bbox_join_region': '.bbox',
    'bbox_inter': '.bbox',
    'bbox_inter_area_matrix': '.bbox',
    'bbox_lurd2xywh': '.bbox',
    'bbox_xywh2lurd': '.bbox',
    'bbox_luwh2xywh': '.bbox',
    'bbox_xywh2luwh': '.bbox',
    'bbox_lurd2luwh': '.bbox',
    'bbox_luwh2lurd': '.bbox',
    'ipynb2pycode': '.ipynb2pycode',
}


def __getattr__(name):
    if name in my_modules:
        module = importlib.import_module(my_modules[name], __package__)
        return getattr(module, name)
    else:
        raise ModuleNotFoundError(f'The import name {name} not in this utils, check if want to import {list(my_modules.keys())}')


__all__ = list(my_modules)



from collections import defaultdict

import multiresolutionimageinterface as mir
import numpy as np
import traceback


class SlideWriter:
    def __init__(self, output_path: str, tile_size: int, dimensions: tuple, spacing: float, **options: str):
        self.output_path = output_path
        self.tile_size = tile_size
        self.W, self.H = dimensions
        # 要求横纵分辨率一致
        self.spacing = spacing

        # 以下进入准备部分
        self._writer = mir.MultiResolutionImageWriter()
        self._writer.openFile(self.output_path)

        # 可以在这个接口类下找到各种各样的可写类型
        # https://github.com/computationalpathologygroup/ASAP/blob/develop/multiresolutionimageinterface/MultiResolutionImageWriter.h
        # self._writer.setDownsamplePerLevel(4)
        # self._writer.setMaxNumberOfPyramidLevels(3)
        # color_type: Monochrome, RGB, RGBA, Indexed

        options = defaultdict(lambda: None, **options)
        color_type = options['color_type'] or 'MONOCHROME'
        data_type = options['data_type'] or 'UCHAR'
        compression = options['compression'] or 'LZW'
        interpolation = options['interpolation'] or 'NEAREST'

        self._writer.setColorType(color_type_map[color_type.upper()])
        self._writer.setDataType(data_type_map[data_type.upper()])
        self._writer.setCompression(compression_map[compression.upper()])
        self._writer.setInterpolation(interpolation_map[interpolation.upper()])

        # 两个版本间存在一些命名不同
        # self._writer.setCompression(mir.LZW)
        # self._writer.setDataType(mir.UChar)
        # self._writer.setInterpolation(mir.NearestNeighbor)
        # # color_type: Monochrome, RGB, RGBA, Indexed
        # color_type = {
        #     'MONOCHROME': mir.Monochrome,
        #     'RGB': mir.RGB,
        #     'RGBA': mir.RGBA,
        #     'INDEXED': mir.Indexed,
        # }[self.color_type.upper()]
        # options = {
        #     "compression": "jpeg",
        #     "jpeg_quality": 90,
        #     "tile_size": 512,
        #     "BigTIFF": True
        # }

        self._writer.setTileSize(self.tile_size)
        self._writer.writeImageInformation(self.W, self.H)
        pixel_size_vec = mir.vector_double()
        pixel_size_vec.push_back(self.spacing)
        pixel_size_vec.push_back(self.spacing)
        self._writer.setSpacing(pixel_size_vec)

    def write(self, tile: np.ndarray, x: int, y: int):
        assert tile.shape[0] == tile.shape[1] == self.tile_size, f'要求写入数与维度数对齐{tile.shape} -- {self.tile_size}'
        self._writer.writeBaseImagePartToLocation(tile.flatten().astype('uint8'), x=int(x), y=int(y))

    def finish(self):
        self._writer.finishImage()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type and not exc_val and not exc_tb:
            self.finish()
        else:
            traceback.print_exc()
        return False


color_type_map = {
    'INVALID': mir.ColorType_InvalidColorType,
    'MONOCHROME': mir.ColorType_Monochrome,
    'RGB': mir.ColorType_RGB,
    'RGBA': mir.ColorType_RGBA,
    'INDEXED': mir.ColorType_Indexed,
}
data_type_map = {
    'INVALID': mir.DataType_InvalidDataType,
    'UCHAR': mir.DataType_UChar,
    'UINT16': mir.DataType_UInt16,
    'UINT32': mir.DataType_UInt32,
    'FLOAT': mir.DataType_Float,
}
compression_map = {
    'RAW': mir.Compression_RAW,
    'JPEG': mir.Compression_JPEG,
    'LZW': mir.Compression_LZW,
    'JPEG2000': mir.Compression_JPEG2000,
}
interpolation_map = {
    'LINEAR': mir.Interpolation_Linear,
    'NEAREST': mir.Interpolation_NearestNeighbor,
}



import os.path
from typing import Union
from pathlib import Path
import numpy as np
from PIL import Image
from .reader_numpy import NumpySlide
from .reader_image import ImageSlide


SlideType = None
try:
    from .reader_openslide import OpenSlide
    SlideType = OpenSlide
except ImportError:
    try:
        from .reader_asap import AsapSlide
        SlideType = AsapSlide
    except ImportError:
        try:
            from .reader_tiff import TiffSlide
            SlideType = TiffSlide
        except ImportError:
            pass


def load(path: Union[str, Path]):
    _, ext = os.path.splitext(path)
    if ext in ('.png', '.jpg', '.jpeg'):
        return ImageSlide(Image.open(path).convert('RGB'))
    if ext in ('.tif', '.svs'):
        if SlideType is None:
            raise ImportError('Lib Imported Failed!')
        return SlideType(path)
    if ext in ('.numpy',):
        return NumpySlide(np.load(path))
    raise TypeError(f'File type not supported! {ext}')



from typing import Tuple
import numpy as np
import cv2


def crop(image: np.ndarray, center: Tuple[float, float], size: Tuple[int, int], degree: float = 0, scale: float = 1, nearest: bool = True) -> np.ndarray:
    """
        切图函数，用于切割给定图像，参数含义如下所示：
        1. 定义一个尺寸为 size 的窗口
        2. 将窗口依 scale 倍数缩放（scale > 1 时窗口变大）
        3. 将窗口旋转 degree 角度（逆时针顺序）
        4. 将窗口中心平移至图像的 center 处
        5. 用窗口在 image 中切取数据，当 nearest 为真时，切取数据均来自 image 原图最近相关点，否则来自邻近点的线性运算
        6. 所采集点回归收拢至原窗口，形成一张尺寸为 size 的图像
        请注意：图像与矩阵的顺逆时针顺序相反，图像逆时针对应矩阵顺时针
    """
    if len(image.shape) == 2:
        return _crop(image, center, size, degree, scale, nearest)
    elif len(image.shape) == 3:
        results = [_crop(image[:, :, i], center, size, degree, scale, nearest) for i in range(image.shape[2])]
        return np.stack(results, axis=2)
    else:
        raise ValueError(f'Shape of image must be array[y, x] or array[y, x, c], but found {type(image)} - {image.shape}')


def _crop(image: np.ndarray, center: Tuple[float, float], size: Tuple[int, int], degree: float = 0, scale: float = 1, nearest: bool = True) -> np.ndarray:
    H, W = image.shape
    x_center, y_center = center
    w, h = map(int, (size, size) if isinstance(size, (int, float)) else size)
    wb = (w - 1) / 2
    hb = (h - 1) / 2
    r = degree * np.pi / 180
    sina, cosa = np.sin(r), np.cos(r)
    w1b = (wb * abs(cosa) + hb * abs(sina)) * scale
    h1b = (hb * abs(cosa) + wb * abs(sina)) * scale
    l1, u1 = round(x_center - w1b), round(y_center - h1b)
    r1, d1 = round(x_center + w1b), round(y_center + h1b)
    # get the img-array -> need to be implement
    img = image[max(0, u1): max(0, d1)+1, max(0, l1): max(0, r1)+1]
    img = np.pad(img, [[max(0, -u1), max(0, d1 - H + 1)], [max(0, -l1), max(0, r1 - W + 1)]])
    if nearest:
        # warp_affine -> https://blog.csdn.net/qq_40939814/article/details/117966835
        # build map-matrix
        x_grid, y_grid = np.meshgrid(np.arange(w), np.arange(h))
        x_grid = x_grid - wb
        y_grid = y_grid - hb
        x_index = ((cosa * x_grid + sina * y_grid) * scale + w1b).round().astype(np.int32)
        y_index = ((cosa * y_grid - sina * x_grid) * scale + h1b).round().astype(np.int32)
        # use the numpy-broadcast
        return img[
            np.clip(y_index, 0, d1 - u1),
            np.clip(x_index, 0, r1 - l1),
        ]
    else:
        ws = round(w1b - wb)
        hs = round(h1b - hb)
        w1b = round(w1b)
        h1b = round(h1b)
        dtype = image.dtype
        # cv2 的 scale 和本函数的 scale 定义相反
        rotation_matrix = cv2.getRotationMatrix2D((w1b, h1b), -degree, 1 / scale)
        img = cv2.warpAffine(img.astype(np.float32), rotation_matrix, (w1b*2+1, h1b*2+1))      # flags=cv2.INTER_NEAREST
        left = ws
        right = ws + w
        up = hs
        down = hs + h
        return img[up: down, left: right].astype(dtype)



from scipy import signal
import cv2
import numpy as np
from skimage import measure
from skimage.measure import regionprops
from copy import deepcopy
from scipy.signal import argrelextrema
from skimage.filters.rank import entropy
from skimage.morphology import disk


def image2mask(image: np.ndarray):
    mask = image.std(axis=2)
    mask = mask > 10
    mask = mask.astype(np.uint8)
    # 消除噪点
    k = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    # k = cv2.getStructuringElement(cv2.MORPH_RECT, (5, 5))
    mask = cv2.dilate(mask, k, dst=mask, iterations=1)
    mask = cv2.erode(mask, k, dst=mask, iterations=2)
    mask = cv2.dilate(mask, k, dst=mask, iterations=1)
    return mask * 255



import numpy as np
import cv2
from numpy.lib.stride_tricks import sliding_window_view
from scipy.ndimage import uniform_filter


def process(image: np.ndarray, s: int = 3, std_thresh: float = 5., b: int = 13, k: int = 17):
    # 手动扩圈
    image = np.pad(image, [(1, 1), (1, 1), (0, 0)])
    image[0, :, :] = image[1, :, :]
    image[-1, :, :] = image[-2, :, :]
    image[:, 0, :] = image[:, 1, :]
    image[:, -1, :] = image[:, -2, :]
    # 在局部计算 std
    patches = sliding_window_view(image, (s, s, 1))
    patches = patches.std(axis=(3, 4, 5))    # 然后在 channel 上对 std 求平均
    mask = patches.mean(axis=2)
    # 再然后均值滤波扩散
    mask = uniform_filter(mask, size=(b, b))
    mask = uniform_filter(mask, size=(b, b))
    mask = uniform_filter(mask, size=(b, b))
    # 再阈值截断二值化
    mask = (mask > std_thresh).astype(np.uint8)
    # 最后做一个腐蚀膨胀
    k = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (k, k))
    mask = cv2.dilate(mask, k, dst=mask, iterations=1)
    mask = cv2.erode(mask, k, dst=mask, iterations=2)
    mask = cv2.dilate(mask, k, dst=mask, iterations=1)
    return mask



import json
import types
from typing import Any, TextIO, BinaryIO, List, Dict, TypeVar, Tuple, Union
from io import TextIOWrapper
from pathlib import Path
from threading import Condition
import numpy as np
from torch import Tensor
import multiprocessing

ndarray = np.ndarray
V = TypeVar('V')

class JassorJsonEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs) -> None: ...
    def encode(self, obj: Any) -> str: ...
    def iterencode(self, o, _one_shot: bool = False): ...

class IOWrapper:
    def __init__(self, write_func: callable, flush_func: callable, close_func: callable): ...

class Logger:
    STEP: int
    DEBUG: int
    INFO: int
    WARNING: int
    def __init__(self, start: int = 0, indentation: int = 1, file: TextIO | IOWrapper | TextIOWrapper | str | Path = ..., con: Condition = None, level: int = 2) -> None: ...
    def close(self) -> None: ...
    def track(self, message: str, prefix: str = ''): ...
    def step(self, message: str): ...
    def debug(self, message: str): ...
    def info(self, message: str): ...
    def warn(self, message: str): ...
    def tab(self): ...
    def __getitem__(self, item: str): ...
    def __enter__(self): ...
    def __exit__(self, exc_type: type | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None): ...

class Merger:
    def __init__(self, temp: ndarray | tuple[int, ...], kernel: ndarray | tuple[int, ...], dtype: type = ..., steep: float = 4, eps: float = 1e-17) -> None: ...
    @staticmethod
    def get_kernel(shape: tuple[int, ...], steep: float) -> ndarray: ...
    def set(self, patch: ndarray, grid: tuple[int | None, ...]) -> None: ...
    def tail(self) -> ndarray: ...

class Process(multiprocessing.Process):
    def __init__(self, target: callable, args=()) -> None: ...
    def start(self) -> None: ...
    def run(self) -> None: ...

class Queue:
    def __init__(self, maxsize: int = 0) -> None: ...
    def ready(self) -> bool: ...
    def is_end(self) -> bool: ...
    def count(self) -> int: ...
    def push(self, item: V) -> None: ...
    def top(self) -> V: ...
    def pop(self) -> V: ...
    def message(self) -> str: ...
    def end(self, flag: bool = False, message: str = '') -> None: ...

class Closed(BaseException):
    def __init__(self, message: str) -> None: ...

class QueueMessageException(BaseException):
    def __init__(self, message: str) -> None: ...
class TimerManager:
    register: dict[str, dict[str, float]]
    def __init__(self, container: str | dict[str, float] = None) -> None: ...
    def __getitem__(self, item: str): ...
    @property
    def costs(self) -> dict[str, float]: ...
    @staticmethod
    def stamp(t: float) -> str: ...
    @staticmethod
    def register_container(key: str, container: dict[str, float]): ...

class Timer:
    def __init__(self, key: str, container: str | dict[str, float]) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type: type | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None): ...
    def __call__(self, func: callable): ...

def random_colors(n: int, start: tuple[int, int, int] = (47, 81, 177), step: tuple[int, int, int] = (73, 119, 31), rand: tuple[int, int, int] = (10, 10, 10)) -> list[tuple[int, int, int]]: ...
def random_rainbow_curves(shape: tuple[int, int, int], s: int = 117, k: int = 7, c: int = 50): ...
def plot(item: Any, title: str = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000, ticks: bool = True): ...
def plots(items: List[Any], titles: List[str] = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000, ticks: bool = True): ...

class Table:
    def __init__(self, *dimensions: List[str] | Dict[str, str], dtype: type = object, data: ndarray = None, key_sep: str = '-', k_v_sep: str = ': ') -> Any: ...
    def __getitem__(self, items) -> Union[Table, object]: ...
    def __setitem__(self, items, value): ...
    def __str__(self) -> str: ...
    @property
    def data(self) -> ndarray: ...
    def dump(self, f: TextIO, type_repr: callable = repr): ...
    def dumpb(self, f: BinaryIO): ...
    def dumps(self, type_repr: callable = repr) -> str: ...
    @staticmethod
    def load(f: TextIO, type_loader: callable = json.loads) -> Table: ...
    @staticmethod
    def loadb(f: BinaryIO) -> Table: ...
    @staticmethod
    def loads(lines: List[str], type_loader: callable = json.loads) -> Table: ...

def uniform_iter(T_len: int, I_len: int, max_gap: int) -> List[int]: ...

def crop(image: ndarray, center: Tuple[float, float], size: Tuple[int, int], degree: float = 0, scale: float = 1, nearest: bool = True) -> ndarray: ...

class SlideWriter:
    # def __init__(self, output_path: str, tile_size: int, dimensions: tuple, spacing: float, **options: str): ...
    def __init__(self, output_path: str, tile_size: int, dimensions: tuple, mpp: float, mag: float,
        level_count: int = 5, interpolation: int = 0, channel: int = 0, dtype: type = np.uint8, **options: str): ...
    # options see:
    # color_type in ['INVALID', 'MONOCHROME', 'RGB', 'RGBA', 'INDEXED']
    # data_type in ['INVALID', 'UCHAR', 'UINT16', 'UINT32', 'FLOAT']
    # compression in ['RAW', 'JPEG', 'LZW', 'JPEG2000']
    # interpolation in ['LINEAR', 'NEAREST']
    def write(self, tile: ndarray, x: int, y: int): ...
    def finish(self): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_val, exc_tb): ...

def image2slide(image: ndarray, out_file: str, mpp: float = 0.5) -> None: ...

class BBox:
    XYWH = 'cxcywh'
    LURD = 'lurd'
    LUWH = 'luwh'
    def __init__(self, bbox: Union[list, ndarray, Tensor], box_format: str = LURD): ...
    @property
    def bbox(self): ...
    @property
    def contour(self): ...
    def join_region(self, l, u, r, d, eps=5): ...
    def area(self): ...
    def inter(self, bbox): ...
    def xywh(self): ...
    def lurd(self): ...
    def luwh(self): ...

def bbox_to_contour(bbox: np.ndarray): ...
def bbox_join_region(bbox: ndarray, l: float, u: float, r: float, d: float, eps: float): ...
def bbox_inter(bbox1: ndarray, bbox2: ndarray): ...
def bbox_inter_area_matrix(bbox1: ndarray, bbox2: ndarray) -> ndarray: ...
def bbox_lurd2xywh(bbox: ndarray) -> ndarray: ...
def bbox_xywh2lurd(bbox: ndarray) -> ndarray: ...
def bbox_luwh2xywh(bbox: ndarray) -> ndarray: ...
def bbox_xywh2luwh(bbox: ndarray) -> ndarray: ...
def bbox_lurd2luwh(bbox: ndarray) -> ndarray: ...
def bbox_luwh2lurd(bbox: ndarray) -> ndarray: ...

def ipynb2pycode(ipynb: str) -> str: ...



import os

import numpy as np
import tifffile
import cv2


'''
svs格式定义，TIFF或BIGTIFF，不能使用subifds
第1张，全分辨率tile图，需设定desc
第2张，缩略图
第3到第N-2张，降分辨率tile图，必须使用从大到小顺序
第N-1张，label图，需设定标志(ReducedImage 1 (0x1))，需设定desc
第N张，marco图，需设定标志 (ReducedImage 1 (0x1), Macro 8 (0x8))，需设定desc
'''


# 一些svs定义
svs_desc = 'Aperio Image Library Fake\nABC |AppMag = {mag}|Filename = {filename}|MPP = {mpp}'
label_desc = 'Aperio Image Library Fake\nlabel {W}x{H}'
macro_desc = 'Aperio Image Library Fake\nmacro {W}x{H}'


def image2slide(image: np.ndarray, out_file: str, mpp: float = 0.5) -> None:

    H, W, _ = image.shape
    w, h = W, H

    # 要需要的金字塔分辨率
    while w > 1000 or h > 1000:
        w //= 2
        h //= 2

    # tile 大小
    t = 64

    # 指定mpp值
    mpp = mpp
    # 指定缩放倍率 (目前为止我们用的都是 0.5 * 20 = 10)
    mag = round(10 / mpp)
    # 换算mpp值到分辨率
    resolution = [10000 / mpp, 10000 / mpp, 'CENTIMETER']
    # 指定图像名字
    filename = os.path.basename(out_file)

    # 尝试写入 svs 格式
    with tifffile.TiffWriter(out_file, bigtiff=True) as tif:

        # outcolorspace 要保持为默认的 YCbCr，不能使用rgb，否则颜色会异常
        # 95 是默认JPEG质量，值域是 0-100，值越大越接近无损
        compression = ['JPEG', 95, dict(outcolorspace='YCbCr')]
        # compression = 'JPEG'
        kwargs = dict(subifds=0, photometric='rgb', planarconfig='CONTIG', compression=compression, dtype=np.uint8, metadata=None)

        tif.write(data=image, shape=(H, W, 3), tile=(t, t), resolution=resolution, description=svs_desc.format(mag=mag, filename=filename, mpp=mpp), **kwargs)
        image = cv2.resize(image, (w, h))
        tif.write(data=image, shape=(h, w, 3), tile=(t, t), resolution=resolution, description='', **kwargs)



import itertools
import json
import pickle
from typing import List, Iterable, Union, Tuple, Dict, TextIO, BinaryIO
import numpy as np


class Table(object):
    """
    这是一个统计表工具
    可以将它理解为 numpy 或 pandas 的扩展（但不使用 pandas）
    它解决的核心痛点是 numpy 不支持 key-word 索引
    考虑一项预测任务，其预测结果对应一族混淆矩阵，这族混淆矩阵可能是高维度数的
    例如：权重序列 weights 对 patches、batches、images 或者 groups 生成了相应的混淆矩阵，混淆矩阵本身又包含 pr、gt 各自类别的两个维度数
    其总体上构成了一个四维张量

    理论上，我们可以用各种各样的方法去存储这种数据结构，包括定义一个形如 [weight, image, pr, gt] 的 numpy 数组
    而问题在于，这个数组本身并不解释其所隶属维度数的含义，每当你切实需要访问其中的一些元素时，你总是需要通过额外的配置信息去确定待访问的索引
    这既麻烦，又容易出错，更不适合数据存储（你需要将相关的数据结构分别存储为配置信息和数据信息，且这二者之间的逻辑关系可能依赖代码去解析，以上三条缺失任意一条都会导致数据破损）

    本工具具备三项功能：
    1. 数据的结构化存储，一个表单就是一套数据，整存整读，不依赖额外配置项或代码逻辑
    2. 表意化索引读写，规避避免复杂的索引转换任务，强化代码可读性
    3. 强化数据可读性，有时候使用代码显示数据会显得很麻烦很啰嗦，本工具支持直接阅读原始数据

    定义一个表单：
    给定维度参数, 可以用字典形式指定相应简写
    axis1 = {'r11': 'resnet-epoch-11', 'e17': 'efficient_net-epoch-17', 'c19': 'custom_net-epoch-19', ...}
    axis2 = ['patch-199', 'batch-21', 'image-7', 'group-train', ...]
    axis3 = {'bg': 'background', 'tg': 'target', 'nt': 'non-target'}
    axis4 = {'bg': 'background', 'tg': 'target', 'nt': 'non-target'}
    table = Table(axis1, axis2, axis3, axis4)

    向表单写入值：
    table[:, :, 'bg', 'bg'] = 0                     # 即： array[:, :, 0, 0] = 0
    table[:, :, ('bg', 'tg'), ('bg', 'tg')] = 0     # 即： array[:, :, 0:2, 0:2] = 0
    table[:, :, 0:2, (0, 1)] = [[0, 0], [0, 0]      # 即： array[:, :, 0:2, 0:2] = 0
    table['r11', 'patch-199'] = 0                   # 即： array[0, 0, :, :] = 0
    table[:, :, 'bg'] = {'bg':0, 'tg':0, 'nt': 0}   # 即： array[:, :, 0, :] = 0

    从表单取值：
    v = table['r11', 'patch-199', 'bg', 'bg']       # int
    v = table['r11', 'patch-199', ('bg', 'tg')]     # table[('bg', 'tg'), ('bg', 'tg', 'ng')]
    v = table['r11', 'patch-199'].array()           # array(3, 3)

    表单存取
    table.set_key_sep('#')  # 不同维度键之间的分割符
    table.set_kv_sep(':')   # 键与数值之间的分割符
    print(table.dumps())    # 直接打印为字符串序列，依赖 sep
    table.dump(TextIO)   # 存成可读文档形式，依赖 sep
    table.dumpb(BinaryIO)  # dumpb 不依赖 sep
    table = Table.loads(table_str)
    table = Table.load(TextIO)
    table = Table.loadb(BinaryIO)

    特别注意的是：在 Table 中, table[0: 2] 视为选取三项：[0, 1, 2], 即，Table 的区间两端包含
    """

    def __init__(self, *dimensions: Union[List[str], Dict[str, str]],
                 dtype: type = object,
                 data: np.ndarray = None,
                 key_sep: str = '-',
                 k_v_sep: str = ': ',):
        self.rank = len(dimensions)
        self.size_in_dims = [len(dimension) for dimension in dimensions]
        self.key_sep = key_sep
        self.k_v_sep = k_v_sep
        self.dim_keys = []
        self.dim_names = []
        for dimension in dimensions:
            if isinstance(dimension, List):
                self.dim_keys.append(dimension)
                self.dim_names.append(dimension)
            else:
                self.dim_keys.append(list(dimension.keys()))
                self.dim_names.append(list(dimension.values()))
        self.key_index_map = [{key: i for i, key in enumerate(dim_key)} for dim_key in self.dim_keys]
        self.name_index_map = [{name: i for i, name in enumerate(dim_name)} for dim_name in self.dim_names]

        self._data = np.empty(shape=self.size_in_dims, dtype=dtype) if data is None else data

    @property
    def data(self) -> np.ndarray:
        return self._data

    def __getitem__(self, items):
        select_indexes_in_dim, grids = self._trans_item_args(items)
        selected_data = self._data[grids]
        if all(isinstance(g, int) for g in select_indexes_in_dim):
            # 四个都是专指，就是取元素值
            return selected_data if self._data.dtype == object else selected_data.item()
        else:
            return self._sub_table(select_indexes_in_dim, selected_data)

    def __setitem__(self, items, value):
        select_indexes_in_dim, grids = self._trans_item_args(items)
        self._data[grids] = value

    def __str__(self):
        return '{\n\t' + self.str(type_repr=repr).replace('\n', '\n\t') + '\n}'

    def dump(self, f: TextIO, type_repr: callable = repr):
        f.write(self.dumps(type_repr=type_repr))

    def dumpb(self, f: BinaryIO):
        dimensions = [
            dict(zip(dim_key, dim_name))
            for dim_key, dim_name in zip(self.dim_keys, self.dim_names)
        ]
        pickle.dump({
            'k-v-sep': self.k_v_sep,
            'key-sep': self.key_sep,
            'dimensions': dimensions,
            'data': self._data,
        }, f)

    def dumps(self, type_repr: callable = repr):
        rank = f'rank[{self.rank}]'
        dtype = f'dtype[{self._data.dtype}]'
        kv_sep = f'k-v-sep[{self.k_v_sep}]'
        key_sep = f'key-sep[{self.key_sep}]'
        dimensions = [
            dict(zip(dim_key, dim_name))
            for dim_key, dim_name in zip(self.dim_keys, self.dim_names)
        ]
        content = self.str(type_repr=type_repr)
        return f'{rank}\n{dtype}\n{kv_sep}\n{key_sep}\n\n' + '\n'.join([json.dumps(dimension) for dimension in dimensions]) + '\n\n' + content

    def str(self, type_repr: callable) -> str:
        dim_keys = [[(index, key) for index, key in enumerate(dim_key)] for dim_key in self.dim_keys]
        cartesian = list(itertools.product(*dim_keys))
        cartesian = [
            (   # indexes, keys
                [item[0] for item in items],
                [item[1] for item in items]
            )
            for items in cartesian
        ]
        lines = [
            f'{self.key_sep.join(keys)}{self.k_v_sep}{type_repr(self._data[tuple(indexes)])}'
            for indexes, keys in cartesian
        ]
        return '\n'.join(lines)

    @staticmethod
    def load(f: TextIO, type_loader: callable = json.loads):
        return Table.loads(f.readlines(), type_loader=type_loader)

    @staticmethod
    def loadb(f: BinaryIO):
        data = pickle.load(f)
        return Table(*data['dimensions'], data=data['data'], k_v_sep=data['k-v-sep'], key_sep=data['key-sep'])

    @staticmethod
    def loads(lines: List[str], type_loader: callable = json.loads):
        head = {}
        for head_pos, line in enumerate(lines):
            line = line.strip()
            if not line: continue
            # 前三行是表头，接下来一行是维度定义，再往后是数据项
            if line.startswith('rank[') and line.endswith(']'):
                head['rank'] = int(line[line.find('[') + 1:-1])
            elif line.startswith('dtype[') and line.endswith(']'):
                head['dtype'] = np.dtype(line[line.find('[') + 1:-1])
            elif line.startswith('k-v-sep[') and line.endswith(']'):
                head['kv_sep'] = line[line.find('[') + 1:-1]
            elif line.startswith('key-sep') and line.endswith(']'):
                head['key_sep'] = line[line.find('[') + 1:-1]
            else:
                head['dimensions'] = [
                    json.loads(lines[pos]) for pos in range(head_pos, head_pos + head['rank'])
                ]
                head['head_pos'] = head_pos
                break
        assert all(v in head for v in ['dtype', 'head_pos', 'kv_sep', 'key_sep', 'rank', 'dimensions'])

        # 现在，是时候加载数据了
        table = Table(
            *head['dimensions'],
            dtype=head['dtype'],
            key_sep=head['key_sep'],
            k_v_sep=head['kv_sep'],
        )
        for line in lines[head['head_pos'] + head['rank']:]:
            line = line.strip()
            if not line: continue
            keys, value = line.split(head['kv_sep'])
            keys = keys.split(head['key_sep'])

            value = type_loader(value)
            table[tuple(keys)] = value
        return table

    def _trans_item_args(self, items):
        # 先将输入规范化为元组
        if not isinstance(items, tuple):
            items = items,
        # 然后补齐缺失轴(检查 Ellipsis)
        if len(items) < self.rank and items[-1] != Ellipsis:
            items = *items, Ellipsis
        if any(item == Ellipsis for item in items):
            pos = items.index(Ellipsis)
            items = *items[:pos], *(slice(None) for _ in range(self.rank - len(items) + 1)), *items[pos+1:]
        # 然后逐项翻译
        select_indexes_in_dim = []
        for dim_index, item in enumerate(items):
            # 若 iterable：逐项翻译
            # 若 slice：按规则翻译
            # 翻译 str： 参照idx
            # 翻译 int： 直取
            if item is None:
                item = slice(None)
            if isinstance(item, int) or isinstance(item, str):
                select = self._trans(dim_index, item, 0)
            elif isinstance(item, Iterable):
                select = self._trans_iter(dim_index, item)
            elif isinstance(item, slice):
                select = self._trans_slice(dim_index, item)
            else:
                raise KeyError(f'Not supported type: {type(item)}')
            # assert bool(select), f'Select nothing is meaningless! Check dim_index[{dim_index}] val[{item}]'
            select_indexes_in_dim.append(select)

        grids = self._select(select_indexes_in_dim)
        return select_indexes_in_dim, grids

    def _trans(self, dim_index: int, item: Union[int, str], default: int) -> int:
        if item is None: return default
        if isinstance(item, int): return item
        if item in self.key_index_map[dim_index]: return self.key_index_map[dim_index][item]
        assert item in self.name_index_map[dim_index], KeyError(f'Key "{item}" not in dim_index-{dim_index}')
        return self.name_index_map[dim_index][item]

    def _trans_slice(self, dim_index: int, item: slice) -> tuple:
        start = self._trans(dim_index, item.start, 0)
        stop = self._trans(dim_index, item.stop, self.size_in_dims[dim_index] - 1)
        step = self._trans(dim_index, item.step, 1)
        return tuple(range(start, stop + 1, step))

    def _trans_iter(self, dim_index: int, items: Iterable) -> tuple:
        return tuple(self._trans(dim_index, item, 0) for item in items)

    @staticmethod
    def _select(select_indexes_in_dim: List[Tuple[int]]) -> Tuple[Union[np.ndarray, int], ...]:
        # 这里只区分两种情况：要么给定一组数值，则使用 meshgrid，要么给定一个数值，则直接留用
        grids: List[np.ndarray] = np.meshgrid(*[indexes for indexes in select_indexes_in_dim if isinstance(indexes, tuple)], indexing='ij')
        next_grids = iter(grids)
        return tuple(next(next_grids) if isinstance(indexes, tuple) else indexes for indexes in select_indexes_in_dim)

    def _sub_table(self, select_indexes_in_dim: List[Tuple[int]], selected_data: np.ndarray):
        keys_list = [
            # 从命名空间中选取相关命名
            [self.dim_keys[dim_index][index] for index in select_indexes]
            for dim_index, select_indexes in enumerate(select_indexes_in_dim)
            # 如果某一个维度里选定了一个确定的维度数，那就没必要保留这个维度的名称了，所以用 tuple 过滤， tuple 表示多选
            if isinstance(select_indexes, tuple)
        ]
        names_list = [
            # 从命名空间中选取相关命名
            [self.dim_names[dim_index][index] for index in select_indexes]
            for dim_index, select_indexes in enumerate(select_indexes_in_dim)
            # 如果某一个维度里选定了一个确定的维度数，那就没必要保留这个维度的名称了，所以用 tuple 过滤， tuple 表示多选
            if isinstance(select_indexes, tuple)
        ]
        return Table(*[dict(zip(keys, names)) for keys, names in zip(keys_list, names_list)], data=selected_data, key_sep=self.key_sep, k_v_sep=self.k_v_sep)



# from .modnet_predict import inference as get_human
# from .classic_fg_sep import none_gray as get_none_gray, none_pure as get_none_pure, key_area as get_key_area

import importlib
my_modules = {
    'get_human': '.modnet_predict',
    'get_none_gray': '.pixel_diff',
    'get_edge': '.area_diff',
    'get_sketch': '.edge_diff',
    'get_valid_area': '.back_free',
}


def __getattr__(name):
    if name in my_modules:
        module = importlib.import_module(my_modules[name], __package__)
        return getattr(module, 'process')
    else:
        raise ModuleNotFoundError(f'The import name {name} not in this utils, check if want to import {list(my_modules.keys())}')


__all__ = list(my_modules)



from scipy import signal
import cv2
import numpy as np


def process(image: np.ndarray, alpha: float = 1., s1: int = 9, s2: int = 5):
    # 过滤非黑白灰区域，适用于显微镜图
    # unify image channels to 3
    if len(image.shape) == 2:
        image = image[:, :, None]
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
    elif image.shape[2] == 4:
        image = image[:, :, 0: 3]
    # 计算结构蒙版
    m1 = np.max(image, 2)
    m2 = np.min(image, 2)
    m3 = m2 / 255 * (255-m1)
    m4 = np.clip(m3, 1, 8)
    k = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (s1, s1))
    m5 = signal.convolve2d(m4, k, mode='same') / k.sum()
    mask = ((m1 - m2) / (m5 * alpha + 1e-19))
    mask = mask.clip(0, 1)
    mask = (mask * 255).round().astype(np.uint8)
    # 捕获边缘
    k = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (s2, s2))
    cv2.dilate(mask, k, dst=mask, iterations=1)
    cv2.erode(mask, k, dst=mask, iterations=2)
    cv2.dilate(mask, k, dst=mask, iterations=1)

    return mask



from typing import List, Tuple

import shapely
from shapely.geometry.base import BaseGeometry

from .definition import Shape, Single, Multi, CoordinatesNotLegalException, NoParametersException
from .impl_base import Base
from . import functional as F


class ComplexPolygon(Base, Single):
    """
    单-复连通多边形, 创建方式有三:
    1. 指定 geo
    2. 指定 single
    3. 指定 outer, *inners
    按上述优先顺序
    """

    __slots__ = ()

    def __init__(
            self,
            outer: List[Tuple[float, float]] = None,
            inners: List[List[Tuple[float, float]]] = None,
            geo: BaseGeometry = None,
            single: Single = None,
            from_p: Tuple[
                List[Tuple[float, float]],          # outer
                List[List[Tuple[float, float]]],    # inners
            ] = None,
            reverse: bool = False
    ):
        if geo is not None:
            assert isinstance(geo, shapely.Polygon), 'geo 必须是 Polygon'
        elif single is not None:
            assert isinstance(single, Single), 'Multi 类型无法转换为 Single'
            geo = single.geo
        elif from_p is not None:
            outer, inners = from_p
            geo = shapely.Polygon(shell=outer, holes=inners)
        elif outer is not None:
            # 对用户输入进行检查和修复
            geo = shapely.Polygon(shell=outer, holes=inners)
            geo = F.norm_geo(geo)
            # 创建时要求轮廓必须合法
            if geo is None or isinstance(geo, shapely.MultiPolygon):
                raise CoordinatesNotLegalException(f'creating single polygon with geo=={type(geo)}')
        else:
            # 没有任何参数的话，要报个错
            raise NoParametersException(f'Any of such parameters have to be provided: (outer, *inners), geo, single, from_p')
        super().__init__(geo=geo, reverse=reverse)

    @property
    def outer(self) -> Single:
        # 外轮廓(正形)
        geo = shapely.Polygon(shell=self.geo.exterior)
        return Single.SIMPLE(geo=geo)

    @property
    def inner(self) -> Multi:
        # 内轮廓(负形)
        inners = [shapely.Polygon(shell=inner) for inner in self.geo.interiors]
        geo = shapely.MultiPolygon(polygons=inners)
        if geo.is_empty:
            return Shape.EMPTY
        return Multi.SIMPLE(geo=geo)

    def sep_in(self) -> Tuple[Single, Multi]:
        # 内分解
        return self.outer, self.inner

    def sep_out(self) -> List[Single]:
        # 外分解
        return [self]

    def sep_p(self) -> Tuple[
        List[Tuple[float, float]],
        List[List[Tuple[float, float]]]
    ]:
        # 点分解
        outer = list(self.geo.exterior.coords)
        inners = [list(inner.coords) for inner in self.geo.interiors]
        return outer, inners

    @property
    def cls(self) -> type:
        return ComplexPolygon


Single.COMPLEX = ComplexPolygon



from typing import Tuple
import numpy as np
from .interface import Reader, num
from PIL import Image
Image.MAX_IMAGE_PIXELS = 16_0000_0000


class ImageSlide(Reader):
    def __init__(self, image: Image.Image, base_mpp: float = 0.5):
        self.image = image
        self._base_mpp = base_mpp

    @property
    def level_count(self) -> int:
        return 1

    @property
    def base_mpp(self) -> float:
        return self._base_mpp

    def dimension(self, level: int = 0) -> Tuple[int, int]:
        return self.image.size

    def downsample(self, level: int = 0) -> float:
        return 1

    def region(self, level: int, left: num, up: num, right: num, down: num) -> np.ndarray:
        patch = self.image.crop((left, up, right, down))
        return np.asarray(patch)



from typing import Tuple
import numpy as np
from .interface import Reader, num


class NumpySlide(Reader):
    def __init__(self, image: np.ndarray, base_mpp: float = 0.5):
        self.image = image
        self.dim = len(self.image.shape)
        self._base_mpp = base_mpp

    @property
    def level_count(self) -> int:
        return 1

    @property
    def base_mpp(self) -> float:
        return self._base_mpp

    def dimension(self, level: int = 0) -> Tuple[int, int]:
        h, w = self.image.shape[:2]
        return w, h

    def downsample(self, level: int = 0) -> float:
        return 1

    def region(self, level: int, left: num, up: num, right: num, down: num) -> np.ndarray:
        left = round(left)
        up = round(up)
        right = round(right)
        down = round(down)
        w, h = self.dimension()
        ml = max(0, left)
        mu = max(0, up)
        mr = min(w, right)
        md = min(h, down)
        patch = self.image[up: down, left: right]
        if ml == left and mu == up and mr == right and md == down:
            return patch.copy()
        patch = np.pad(patch, [(mu-up, down-md), (ml-left, right-mr), (0, 0)][:self.dim])
        return patch



from PIL import Image
from typing import List, Tuple, Any, Union
from pathlib import Path
import numpy as np
import torch
from .interface import Reader
from .reader import load
from .reader_numpy import NumpySlide
from .reader_image import ImageSlide


class SingleDataset(torch.utils.data.Dataset):
    """
    单图预测任务
    """
    def __init__(self, source: Union[str, Path, np.ndarray, Image.Image], samples: List[Tuple[int, int, int, int, int]]):
        super().__init__()
        self.source = source
        # [(level, left, up, right, down)]
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, item):
        return self.load(*self.samples[item])

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def load(self, level: int, left: int, up: int, right: int, down: int) -> Any:
        """
        这里的复杂度在于：
        1. load 方法可能被多进程执行，若数据不共享，则每个子进程需要 copy 数据，而 io 管道不能被 copy
        2. image 与 slide 要采用相同的外部封装，但二者的代码逻辑完全不同
        """
        # 解析路径
        if isinstance(self.source, Reader):
            pass
        elif isinstance(self.source, (str, Path)):
            self.source = load(self.source)
        elif isinstance(self.source, Image.Image):
            self.source = ImageSlide(self.source)
        elif isinstance(self.source, np.ndarray):
            self.source = NumpySlide(self.source)
        else:
            raise TypeError(f'No such type supporting! {type(self.source)}')
        return self.source.region(level=level, left=left, up=up, right=right, down=down)



from typing import Union
from pathlib import Path
import numpy

ndarray = numpy.ndarray

def get_human(image: ndarray, onnx_path: Union[str, Path] = './modnet_human_detect.onnx') -> ndarray: ...

def get_none_gray(image: ndarray, alpha: float = 1., s1: int = 9, s2: int = 5) -> ndarray: ...

def get_edge(image: ndarray, s: int = 9, score_thresh: float = 0.012) -> ndarray: ...

def get_sketch(image: ndarray, blur_size: int = 9, thresh_min: float = 0, thresh_max: float = 1.) -> ndarray: ...

def get_valid_area(image: ndarray, s: int = 3, std_thresh: float = 5., b: int = 13, k: int = 17) -> ndarray: ...



import inspect
from skimage.transform import resize
import os
import traceback
from pathlib import Path
import zstandard as zstd
import numpy as np
import tifffile


class SlideWriter:
    """
    svs格式定义，TIFF或BIGTIFF，不能使用subifds
    第1张，全分辨率tile图，需设定desc
    第2张，缩略图
    第3到第N-2张，降分辨率tile图，必须使用从大到小顺序
    第N-1张，label图，需设定标志(ReducedImage 1 (0x1))，需设定desc
    第N张，marco图，需设定标志 (ReducedImage 1 (0x1), Macro 8 (0x8))，需设定desc
    """

    def __init__(
            self,
            output_path: str,
            tile_size: int,
            dimensions: tuple,
            mpp: float,
            mag: float,
            compression: str = 'LZW',
            photometric: str = 'minisblack',
            level_count: int = 5,
            name: str = '',
            # interpolation: int = cv2.INTER_NEAREST,
            interpolation: str = 'Nearest',
            channel: int = 0,
            dtype: type = np.uint8,
            **options: str
    ):
        self.output_path = output_path
        self.tile_size = tile_size
        self.W, self.H = dimensions
        # 要求横纵分辨率一致
        self.mpp = mpp
        self.mag = mag
        self.level_count = level_count
        self.interpolation = interpolation_map[interpolation.upper()]   # 插值方式选项
        self.compression = compression_map[compression.upper()]  # 压缩方式选项
        self.photometric = photometric_map[photometric.upper()]  # 颜色选项
        self.options = options
        self.channel = channel
        self.dtype = dtype
        self.shapes = [
            (self.H // 2 ** level, self.W // 2 ** level)
            if self.channel == 0 else
            (self.H // 2 ** level, self.W // 2 ** level, self.channel)
            for level in range(level_count)
        ]
        self.tile_shapes = [
            (self.tile_size // 2 ** level, self.tile_size // 2 ** level)
            if self.channel == 0 else
            (self.tile_size // 2 ** level, self.tile_size // 2 ** level, self.channel)
            for level in range(level_count)
        ]
        self.tile_shape = self.tile_shapes[0]
        self.name = name or Path(output_path).name
        self.desc = f'Aperio Image Library Fake\nABC |AppMag = {self.mag}|Filename = {self.name}|MPP = {self.mpp}'
        self.options = dict(
            subifds=0, dtype=self.dtype,
            photometric=self.photometric, compression=self.compression,
            planarconfig='CONTIG', metadata=None,
            **options
        )

        # 这个必须按流式方法，一次性写入，我需要手动管理一个类似 ASAP 写图时的缓冲区的东西
        self._writer = tifffile.TiffWriter(output_path, bigtiff=True)
        self._buffer_path = f'{output_path}.buffer'
        self._buffer = open(self._buffer_path, 'wb')
        self._info_cache = {}
        self.cctx = zstd.ZstdCompressor()
        self.dctx = zstd.ZstdDecompressor()

        # 检查参数是否符合匹配要求，不符合直接报错
        try:
            sig = inspect.signature(self._writer.write)
            sig.bind(data=None, shape=None, tile=None, description=None, **self.options)  # 模拟 func(**options) 的参数匹配
        except TypeError as e:
            print(f"参数不匹配: {e}")
            raise e

    def write(self, tile: np.ndarray, x: int, y: int):
        assert tile.shape == self.tile_shape, f'要求写入数与维度数对齐{tile.shape} -- {self.tile_shape}'

        # 分层级存储
        info = []
        for level in range(self.level_count):
            tile = resize(tile, self.tile_shapes[level][:2], order=self.interpolation)
            # tile = cv2.resize(tile, self.tile_shapes[level][:2], interpolation=self.interpolation)
            # 编码
            # success, buffer = cv2.imencode('.png', tile)
            # if not success: raise RuntimeError("图像压缩失败")
            # 使用 zstd 压缩格式
            buffer = self.cctx.compress(tile.flatten().tobytes())
            # 写入缓冲区
            pos_start = self._buffer.tell()
            self._buffer.write(buffer)  # 写入 PNG 数据
            pos_end = self._buffer.tell()
            info.append((pos_start, pos_end))
        # 记录写图信息
        self._info_cache[(x, y)] = info

    def finish(self):
        self._buffer.close()
        self._buffer = open(self._buffer_path, "rb")
        try:
            # 第 1 张，完整全图，带描述
            stream = self.load_buffer(level=0)
            self._writer.write(data=stream, shape=self.shapes[0], tile=self.tile_shapes[0][:2], description=self.desc, **self.options)
            # 降分辨率tile图，从大到小
            for level in range(1, self.level_count):
                stream = self.load_buffer(level=level)
                self._writer.write(data=stream, shape=self.shapes[level], tile=self.tile_shapes[level][:2], description='', **self.options)
        except Exception as e:
            traceback.print_exc()
        finally:
            # 完毕
            self._writer.close()
            self._buffer.close()
            os.remove(self._buffer_path)

    def load_buffer(self, level: int):
        space = np.ones(self.tile_shapes[0], self.dtype)
        H, W = self.shapes[0][:2]
        for y in range(0, H, self.tile_size):
            for x in range(0, W, self.tile_size):
                if (x, y) not in self._info_cache:
                    yield space
                    continue
                pos_start, pos_end = self._info_cache[(x, y)][level]
                self._buffer.seek(pos_start)
                buffer = self._buffer.read(pos_end - pos_start)
                image_bytes = self.dctx.decompress(buffer)
                img = np.frombuffer(image_bytes, dtype=self.dtype).reshape(self.tile_shapes[level])
                # img = cv2.imdecode(img_array, cv2.IMREAD_UNCHANGED)
                yield img

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type and not exc_val and not exc_tb:
            self.finish()
        else:
            traceback.print_exc()
            try:
                self._writer.close()
            except Exception as e:
                pass
            if not self._buffer.closed: self._buffer.close()
        return False


photometric_map = {
    # 来自 GPT
    'MINISWHITE': 0,  # 单通道图像中，0 表示白色，最大值表示黑色（老式习惯）
    'MINISBLACK': 1,  # 单通道图像中，0 表示黑色，最大值表示白色（现代标准）
    'RGB': 2,  # 多通道彩色图像，3 个通道分别表示 R、G、B
    'PALETTE': 3,  # 图像值为颜色查找表（colormap）中的索引
    'MASK': 4,  # 通常用于蒙版或 alpha
    'SEPARATED': 5,  # CMYK 色彩空间（青品黄黑）
    'YCBCR': 6,  # YCbCr 色彩空间（通常用于 JPEG 编码）
    'CIELAB': 8,  # Lab 色彩空间
    'CFA': 32803,  # Color Filter Array，传感器原始图像（如 Bayer 图）
    'LOGHUFFMAN': 32844,  # 用于 LogLuv 图像
    'LINEARRAW': 34892,  # 用于 Raw 图像数据
    # 在 tifffile 中还可能支持字符串形式，比如：'minisblack'、'rgb'、'palette'、'cfa' 等。
}
compression_map = {
    'NONE': 1,  # 单通道图像中，0 表示白色，最大值表示黑色（老式习惯）
    'LZW': 5,  # Lempel-Ziv-Welch 压缩
    'JPEG': 6,  # JPEG 压缩（通常与 YCbCr 配合）
    'JPEGOLD': 7,  # 老式 JPEG（不推荐）
    'ZLIB': 8,  # zlib 压缩
    'DEFLATE': 32946,  # ADOBE 定义的 deflate 算法
    'PACKBITS': 32773,  # 一种简单的 RLE 压缩
    'LERC': 34925,  # Limited Error Raster Compression，用于 GIS
    'ZSTD': 50001,  # Facebook 提出的压缩算法
    'LZMA': 50002,  # Lempel-Ziv-Markov 压缩
    'WEBP': 50003,  # 用于 Web 图像压缩
    'auto': 'auto',  # 自动选择（仅在某些 tifffile 函数中支持）
}
interpolation_map = {
    'NEAREST': 0,
    'BILINEAR': 1,
    'LINEAR': 1,
}



from scipy import signal
from scipy import ndimage
import cv2
import numpy as np


def process(image: np.ndarray, s: int = 9, score_thresh: float = 0.012):
    # 探测边缘（我也不知道为什么，但总之，它确实可以做到）
    # unify image channels to 3
    if len(image.shape) == 2:
        image = image[:, :, None]
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
    elif image.shape[2] == 4:
        image = image[:, :, 0: 3]
    image = image / 255
    diff: np.ndarray = np.zeros_like(image)
    # 中值滤波对均值滤波做差
    k = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (s, s))
    for i in range(3):
        mean_filter = signal.convolve2d(image[..., i], k, mode='same') / k.sum()
        mid_filter = ndimage.median_filter(image[..., i], size=s, mode='reflect')
        diff[..., i] = abs(mean_filter - mid_filter)
    # 差主体低于 0.1 的视为背景板
    mask = diff.max(axis=2) / score_thresh
    # mask = diff.sum(axis=2) > score_thresh
    mask = mask.clip(0, 1)
    mask = (mask * 255).round().astype(np.uint8)

    return mask
